#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

This plugins is a combination of two plug-ins Adrian Likins ( http://www.adrianlikins.com/misc/gimp/ ): font-to-brushes.py and font-to-image-hose.py

Attention: Do not use these characters \/?:*.><| to save the as file
'''

from gimpfu import *
import unicodedata
import math
import sys,os,os.path,re,glob,traceback
from collections import namedtuple

#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   The GPL v3 licence is available at: https://www.gnu.org/licenses/gpl-3.0.en.html

# ------------
#| Change Log |
# ------------
# Rel 0.1: Initial release. Code embedded: ofn_tiles_to_layer.py by Ofnuts
# Rel 0.2: New options: Add a keyboard key?, national charters by Skinnyhouse ("French","German","Italian","Polish","Spanish")


################################################ Start ofn_tiles_to_layer ########################################################################################################################
debug=False
def trace(s):
    if debug:
        print s

def createOpts(name,pairs):
    optsclass=namedtuple(name+'Type',[symbol for symbol,label in pairs]+['labels','labelTuples'])
    opts=optsclass(*(
                    range(len(pairs))
                    +[[label for symbol,label in pairs]]
                    +[[(label,i) for i,(symbol,label) in enumerate(pairs)]]
                    ))
    return opts

JoinDefs=createOpts('JoinDefs',[('ROWS_GIVEN','Rows given, compute columns\nPodane wiersze, oblicz kolumny'),('COLUMNS_GIVEN','Columns given, compute rows\nPodane kolumny, oblicz wiersze'),('GUESS','Parse tile names\nAnalizuj nazwy kafelków')])
JoinStackOrder=createOpts('JoinStackOrder',[('FROM_TOP','From top\nZ góry'),('FROM_BOTTOM','From bottom\nZ dołu')])
JoinFlatten=createOpts('JoinFlatten',[('NO','Keep as distinct layers\nZachowaj jako odrębne warstwy'),('YES','Flatten to single layer\nSpłaszcz do pojedynczej warstwy')])
RCOrder=createOpts('RCOrder',[('ROWS_FIRST','Rows first\nNajpierw rzędy'),('COLUMNS_FIRST','Columns first\nNajpierw kolumny')])
Resize=createOpts('Resize',[('NO','No change\nBrak zmiany'),('YES','Fit to result\nDopasuj do wyniku')])
SplitInPlace=createOpts('SplitInPlace',[('IN_PLACE','Keep in place\nPrzytrzymać'),('SUPERIMPOSE','Superimpose\nNakładać')])

def iterateTiles(rows,columns,rowsFirst):
    if rowsFirst:
        for row in range(rows):
            for column in range(columns):
                yield column*rows+row,row*columns+column,row,column
    else:
        for column in range(columns):
            for row in range(rows):
                yield column*rows+row,row*columns+column,row,column
            
def setValues(layerName,rows,columns,width,height,row,column,tileCR,tileRC):
    formatValues={}
    formatValues['layerName']=layerName
    formatValues['rows']=rows
    formatValues['columns']=columns
    formatValues['width']=width
    formatValues['height']=height
    formatValues['count']=rows*columns
    formatValues['row0']=row
    formatValues['row1']=row+1
    formatValues['column0']=column
    formatValues['column1']=column+1
    formatValues['tileCR0']=tileCR
    formatValues['tileCR1']=tileCR+1
    formatValues['tileRC0']=tileRC
    formatValues['tileRC1']=tileRC+1
    return formatValues
        
def layerToTiles(image,layer,namePattern,tileData,inPlace,resize,rcOrder):
    image.undo_group_start()
    
    try:
        if type(layer) is not gimp.Layer:
            raise Exception('Source element must be a plain layer')
        width,height,rows,columns=tileData
        for tileCR,tileRC,row,column in iterateTiles(rows,columns,rcOrder==RCOrder.ROWS_FIRST):
            formatValues=setValues(layer.name,rows,columns,width,height,row,column,tileCR,tileRC)
            try:
                tileName=namePattern.format(**formatValues)
            except KeyError as e:
                raise Exception('No formatting variable called "%s"' % e.args[0])        
            tile=layer.copy()
            image.add_layer(tile,0)
            tile.resize(width,height,-column*width,-row*height)
            tile.name=tileName
            if inPlace==SplitInPlace.SUPERIMPOSE:
                tile.set_offsets(0,0)
            trace('Tile W*H: %d*%d' % (tile.width,tile.height))
        image.remove_layer(layer)
        if inPlace==SplitInPlace.SUPERIMPOSE and resize==Resize.YES:
            image.resize(width,height)

    except Exception as e:
        pdb.gimp_message(e.args[0])
        print traceback.format_exc()
    
    image.undo_group_end()

def tilesGuessedFromNames(layers,rcOrder):
    pattern=re.compile(r'(^(.*\D)?)(\d+)(\D+)(\d+)(\D*)$')
    commonSeps=None
    tilesCoords=[]
    for layer in layers:
        match=pattern.match(layer.name)
        if not match:
            raise Exception('Cannot figure out row or columns in layer "%s"' % layer.name)
        s1,_,n1,s2,n2,s3=match.groups()
        if not commonSeps: # first image
            commonSeps=(s1,s2,s3)
        else:
            if commonSeps!=(s1,s2,s3):
                raise Exception('Layer "%s" doesn\'t match %s**%s**%s' % ((layer.name,)+commonSeps))
        if rcOrder==RCOrder.ROWS_FIRST:
            tc=(int(n2),int(n1),layer)
        else:
            tc=(int(n1),int(n2),layer)
        tilesCoords.append(tc)

    minCol=min([x for x,y,l in tilesCoords])
    maxCol=max([x for x,y,l in tilesCoords])
    minRow=min([y for x,y,l in tilesCoords])
    maxRow=max([y for x,y,l in tilesCoords])
    rows=1+maxRow-minRow
    cols=1+maxCol-minCol
    count=rows*cols
    trace('Guessed: %d rows and %d columns' % (rows,cols))
    if count!=len(tilesCoords):
        raise Exception('Number of layers (%d) does not match the number of expected tiles (%d*%d=%d)' % (len(layers),rows,cols,count))
    tiles=[[None for _ in range(rows)] for _ in range(cols)]
    for x,y,l in tilesCoords:
        tiles[x-minCol][y-minRow]=l
    return tiles,rows,cols

def tilesFromGivenRC(layers,defs,geometry,rcOrder):
    if defs== JoinDefs.ROWS_GIVEN:
        rows,cols=geometry,len(layers)/geometry
    else:
        rows,cols=len(layers)/geometry,geometry
        
    if len(layers)!=rows*cols:
        raise Exception('Number of layers (%d) does not match the number of expected tiles (%d*%d=%d)' % (len(layers),rows,cols,rows*cols))

    if rcOrder==RCOrder.ROWS_FIRST:
        tiles=[layers[x::cols] for x in range(cols)]
    else:
        tiles=[layers[x:x+rows] for x in range(0,len(layers),rows)]
        
    return tiles,rows,cols

def dumpTiles(tiles):
    if not debug:
        return
    trace(tiles)
    for y in range(len(tiles[0])): 
        print ' '.join([l.name.strip() for l in [tiles[x][y] for x in range(len(tiles))]])

def tilesToLayerUnicode(image,defs,geometry,rcOrder,stackOrder,flatten,resize):
    image.undo_group_start()
    try:
        layers=[l for l in image.layers if type(l) is gimp.Layer and l.visible]
        if stackOrder==JoinStackOrder.FROM_BOTTOM:
            layers.reverse()
        if defs==JoinDefs.GUESS:
            tiles,rows,columns=tilesGuessedFromNames(layers,rcOrder)
        else:
            tiles,rows,columns=tilesFromGivenRC(layers,defs,int(geometry),rcOrder)
        
        dumpTiles(tiles)

        tileW,tileH=tiles[0][0].width,tiles[0][0].height
        for y in range(rows):
            for x in range(columns):
                tile=tiles[x][y]
                if (tileW,tileH)!=(tile.width,tile.height):
                        raise Exception('Tile "%s" is not the same size as the previous ones' % tile.name)
                tile.set_offsets(x*tileW,y*tileH)
            
        if resize==Resize.YES:
            image.resize(columns*tileW,rows*tileH)
        if flatten==JoinFlatten.YES: 
            tilesLayer=image.merge_visible_layers(EXPAND_AS_NECESSARY)
            tilesLayer.name='Tiles'
        
    except Exception as e:
        pdb.gimp_message(e.args[0])
        print traceback.format_exc()
    image.undo_group_end()
	
################################################ End ofn_tiles_to_layer ########################################################################################################################
	
def python_fu_font_table_display(image, layer, font, size, color_f, add_key, color_t, chars, add_nat, iter, h_font, h_font_size, h_color, margin, merge):
	print font, chars, size
	chars = chars.decode("utf-8")
	parts = font.split(" ")
	if parts[-1].isdigit():
		fontparts  = parts[:-1]
		font = string.join(fontparts, " ")

	max_width = 0.0
	max_height = 0.0
	for char in chars:
		(width, height, asc, dsc) = pdb.gimp_text_get_extents_fontname(char, size, 1, font)

		if width > max_width:
			max_width = width
		if height > max_height:
			max_height = height

	image = gimp.Image(max_width*2, max_height*2, RGB)

	layer_count = 0
	drawables = []
	new_max_height = 0

	#  start national charters by Skinnyhouse.##################################################################################
	
	if add_nat==1:
		french_case = unichr(192) + unichr(194) + unichr(196) + unichr(198) + unichr(199) + unichr(200) + unichr(201) + unichr(202) + unichr(203) + unichr(206) + unichr(207) + unichr(212) + unichr(217) + unichr(219) + unichr(220) + unichr(226) + unichr(228) + unichr(230) + unichr(231) + unichr(232) + unichr(233) + unichr(234) + unichr(235) + unichr(238) + unichr(239) + unichr(249) + unichr(251) + unichr(252)
		chars += french_case	
	if add_nat==2:
		german_case = unichr(196) + unichr(214) + unichr(220) + unichr(223) + unichr(228) + unichr(246) + unichr(252)
		chars += german_case
	if add_nat==3:		
		italian_case = unichr(192) + unichr(200) + unichr(201) + unichr(204) + unichr(210) + unichr(217) + unichr(224) + unichr(232) + unichr(233) + unichr(236) + unichr(242) + unichr(249)
		chars += italian_case
	if add_nat==4:	
		polish_case = unichr(260) + unichr(262) + unichr(280) + unichr(321) + unichr(323) + unichr(211) + unichr(346) + unichr(377) + unichr(379) + unichr(261) + unichr(263) + unichr(281) + unichr(322) + unichr(324) + unichr(243) + unichr(347) + unichr(378) + unichr(380)         
		chars += polish_case
	if add_nat==5:		
		spanish_case = unichr(161) + unichr(191) + unichr(193) + unichr(201) + unichr(205) + unichr(209) + unichr(211) + unichr(218) + unichr(220) + unichr(225) + unichr(233) + unichr(237) + unichr(241) + unichr(243) + unichr(250) + unichr(252)	
		chars += spanish_case
		
	#  end national charters by Skinnyhouse.######################################################################################
	
	for char in chars:
		(width, height, asc, dsc) = pdb.gimp_text_get_extents_fontname(char, size, 1, font)
		#Tin Tran: here's a bug with WideAwakeBlack.ttf reported by Wallace.
		#So here we're calling for max_width and max_height instead of width and height then we'll crop it and resize it later.
		#The wideawakeblack.ttf is just a mess with width not returning properly.
		drawable = gimp.Layer(image, "%s" % char, max_width*2, max_height*2, RGBA_IMAGE, 100, NORMAL_MODE)
		image.add_layer(drawable, layer_count)
		layer_count = layer_count + 1
		pdb.gimp_edit_clear(drawable)
		
		font_layer = pdb.gimp_text_fontname(image, drawable,  0.0, 0.0, char, 0, 1, size, 1, font)
		
		pdb.gimp_text_layer_set_color(font_layer, color_f)
		pdb.gimp_floating_sel_anchor(font_layer)
		#pdb.plug_in_autocrop_layer(img,drawable)
		
		if add_key==1:
			name = pdb.gimp_layer_get_name(drawable)		
			text_layer = pdb.gimp_text_fontname(image, drawable, 20,-5 - (size/3), name, 0, 1, size/3, 1, "Arial")

			pdb.gimp_text_layer_set_color(text_layer, color_t)
			
			pdb.gimp_floating_sel_anchor(text_layer)
		#pdb.plug_in_autocrop_layer(img,drawable)

		w = pdb.gimp_image_width(image)
		h = pdb.gimp_image_height(image)
		wh = pdb.gimp_drawable_width(drawable)
		hh = pdb.gimp_drawable_height(drawable)		
		centerx=w/2
		adw=wh/2
		sw=centerx - adw		
		pdb.gimp_layer_set_offsets(drawable, sw, 0)	
		pdb.gimp_layer_resize_to_image_size(drawable)

		pdb.plug_in_autocrop_layer(image,drawable)
		offsetx,offsety = pdb.gimp_drawable_offsets(drawable)
		drawables.append(drawable)
		new_height = offsety + drawable.height
		if new_height > new_max_height:
			new_max_height = new_height
	for drawable in drawables:
		offsetx,offsety = pdb.gimp_drawable_offsets(drawable)
		pdb.gimp_layer_resize(drawable,drawable.width,new_max_height,0,offsety)
	pdb.gimp_image_resize_to_layers(image)
	
	for drawable in drawables:
		pdb.gimp_layer_resize_to_image_size(drawable)	
	name = "%s%s"  % ("", "")
	filename = "%s%s" % ("", "")

	#print img, drawable, filename, filename, 50, name, max_width, new_max_height, 1, 1, 1, [len(chars)], 1
	disp = gimp.Display(image)
	
	##############################################################################
	##############################################################################
	
	for layer in image.layers:
		wi = pdb.gimp_image_width(image)
		hi = pdb.gimp_image_height(image)
		
		new_width = wi + 2*margin
		new_height = hi + 2*margin
		
		wh = pdb.gimp_drawable_width(layer)
		hh = pdb.gimp_drawable_height(layer)
		centerx=wi/2
		centery=hi/2
		adw=wh/2
		adh=hh/2
		sw=centerx - adw
		sh=centery - adh		
		pdb.gimp_layer_set_offsets(layer, sw, sh)
		pdb.gimp_layer_resize(layer,new_width,new_height,margin, margin)
	

# add border
	wh = pdb.gimp_drawable_width(layer)
	hh = pdb.gimp_drawable_height(layer)
	for layer in image.layers:
		pdb.gimp_image_resize_to_layers(image)
		selection = pdb.gimp_item_is_selection(layer)
		pdb.gimp_selection_invert(image)
		pdb.gimp_image_select_rectangle(image, 2, 4, 4, wh-8, hh-8)
		pdb.gimp_selection_invert(image)
		pdb.gimp_image_select_rectangle(image, 3, 2, 2, wh-4, hh-4)
		pdb.gimp_context_set_foreground((0,0,0))
		pdb.gimp_edit_fill(layer, 0)
		pdb.gimp_selection_none(image)				
# create tiles
	#pdb.gimp_image_remove_layer(image, layer)
	wrkimg=gimp.image_list()[0]
	wrklay=wrkimg.layers[-1]
	man = len(image.layers)  
	many = len(image.layers)/iter  
	ile = math.ceil(many)   
	if man != ile*iter:

		dup = (ile)*iter - man			# dup = (ile+1)*iter - man
		for item in range(0,int(dup)):
			
			copy=pdb.gimp_layer_copy(wrklay,1)
			new = pdb.gimp_image_add_layer(image,copy,-1)
			pdb.gimp_image_lower_item_to_bottom(image, copy)
			pdb.gimp_edit_clear(copy)

	#pdb.python_fu_ofn_tiles_to_layer(image, specs, geometry, rcOrder, stackOrder, flatten, resize)
	#pdb.python_fu_ofn_tiles_to_layer(image, 1, iter, 0, 0, 0, 1)
	tilesToLayerUnicode(image, 1, iter, 0, 0, 0, 1)
	pdb.gimp_context_set_foreground((0,0,0))
	
	rkimg=gimp.image_list()[0]
	rklay=rkimg.layers[-1]
	pdb.gimp_layer_resize_to_image_size(rklay)

	wi = pdb.gimp_image_width(image)
	hi = pdb.gimp_image_height(image)
	margin =20
	new_width = wi + 2*margin
	new_height = hi + 2*margin
	
	wh = pdb.gimp_drawable_width(rklay)
	hh = pdb.gimp_drawable_height(rklay)
	centerx=wi/2
	centery=hi/2
	adw=wh/2
	adh=hh/2
	sw=centerx - adw
	sh=centery - adh		
	pdb.gimp_layer_set_offsets(rklay, sw, sh)
	pdb.gimp_image_resize(image,new_width,new_height,margin, margin)
	pdb.gimp_layer_resize_to_image_size(rklay)
	if man == ile*iter:
		layer = pdb.gimp_layer_new(image, new_width,new_height, RGBA_IMAGE, "background",100, 0)
		pdb.gimp_image_insert_layer(image, layer, None, 0)
		pdb.gimp_edit_fill(layer, 2)
		pdb.gimp_image_lower_item_to_bottom(image, layer)
	if man != ile*iter:			
		pdb.gimp_edit_fill(rklay, 2)
		pdb.gimp_context_set_foreground((0,0,0))
	ding = font	
	heading_layer = pdb.gimp_text_fontname(image, None, 0, -h_font_size-margin, font, margin, 1, h_font_size, PIXELS, h_font)
	pdb.gimp_text_layer_set_color(heading_layer, h_color)
	pdb.gimp_image_resize_to_layers(image)
	back=gimp.image_list()[0]
	bl=back.layers[-1]	
	pdb.gimp_layer_resize_to_image_size(bl)
	pdb.gimp_edit_fill(bl,2)
	if merge:
		pdb.gimp_image_merge_visible_layers(image, 0)	
	return
	
register(
	"python_fu_font_table_display",
	"Font Table or Dingbat Symbol Display",
	"Font to layers: Symbol Display V-0.2",
	"MareroQ",
	"GPL",
	"2020",
	"Font Symbol Display...",
	"*",
	[
	(PF_IMAGE,    "image", "Input image", None), #current image  
	(PF_DRAWABLE, "layer", "Input layer", None), #current layer
	(PF_FONT,     "font", "Font Table: ", "Zapf Dingbats Bt"),
	(PF_FLOAT,    "size", "Size of Font: ", 100.0),
	(PF_COLOR,    "color_f", "Font Color:", (0, 0, 0)),
	(PF_TOGGLE,   "add_key",   "Add a keyboard key\n(for font dingbat):", 1),
	(PF_COLOR,    "color_t", "Key Color:", (30,20,150)),	
	(PF_STRING,   "text", "Characters: ", "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ\/?:;*.><|~!@#$%^&()_+-=',{}[]"),
	(PF_OPTION,   "add_nat", "Add National Charters", 0,("No","French","German","Italian","Polish","Spanish")),	
	(PF_SPINNER,  "iter", "Number of Tile Columns:", 10, (2, 30, 1)),
	(PF_FONT,     "h_font", "Heading - Font Name:", "Arial"),	
	(PF_SPINNER,  "h_font_size", "Heading - Font Size:", 50, (10, 500, 1)),
	(PF_COLOR,    "h_color", "Heading - Font Color:", (0, 0, 0)),
	(PF_SPINNER,  "margin", "Add Margin x=y (px): ", 20, (0, 5000, 1)),
	(PF_BOOL,     "merge",  "Merge layers:", False),	
	],
	[],
	python_fu_font_table_display,
	menu="<Image>/Filters/Text/",
	)

main()